@@ -1,4 +1,5 @@
 import logging
+import os.path
 import sys
 import pytest
 
@@ -20,13 +21,9 @@ def test_no_report(misc_testdir):
 
 
 def test_create_report(misc_testdir):
-    res = misc_testdir.runpytest('--json-report')
+    misc_testdir.runpytest('--json-report')
     assert (misc_testdir.tmpdir / '.report.json').exists()
 
-    res.stdout.fnmatch_lines([
-        '*report written*.report.json*',
-    ])
-
 
 def test_create_report_file_from_arg(misc_testdir):
     misc_testdir.runpytest('--json-report', '--json-report-file=arg.json')
@@ -34,10 +31,32 @@ def test_create_report_file_from_arg(misc_testdir):
 
 
 def test_create_no_report(misc_testdir):
+    misc_testdir.runpytest('--json-report', '--json-report-file=NONE')
+    assert not (misc_testdir.tmpdir / '.report.json').exists()
+
+
+def test_terminal_summary(misc_testdir):
+    res = misc_testdir.runpytest('--json-report')
+    res.stdout.fnmatch_lines(['-*JSON report*-', '*report saved*.report.json*'])
+
+    res = misc_testdir.runpytest('--json-report', '--json-report-file=./')
+    res.stdout.fnmatch_lines(['*could not save report*'])
+
     res = misc_testdir.runpytest('--json-report', '--json-report-file=NONE')
-    res.stdout.fnmatch_lines([
-        '*no JSON report written*',
-    ])
+    res.stdout.no_fnmatch_line('-*JSON report*-')
+
+    res = misc_testdir.runpytest('--json-report', '--json-report-file=NONE', '-v')
+    res.stdout.fnmatch_lines(['*auto-save skipped*'])
+
+    res = misc_testdir.runpytest('--json-report', '-q')
+    res.stdout.no_fnmatch_line('-*JSON report*-')
+
+    res = misc_testdir.runpytest('--json-report', '-q', '--json-report-verbosity=0')
+    res.stdout.fnmatch_lines(['-*JSON report*-'])
+
+    res = misc_testdir.runpytest(
+        '--json-report', '--json-report-file=NONE', '-vv', '--json-report-verbosity=0')
+    res.stdout.no_fnmatch_line('-*JSON report*-')
 
 
 def test_report_keys(num_processes, make_json):
@@ -52,7 +71,7 @@ def test_report_keys(num_processes, make_json):
     assert set(data) == keys
     assert isinstance(data['created'], float)
     assert isinstance(data['duration'], float)
-    assert data['root'].startswith('/')
+    assert os.path.isabs(data['root'])
     assert data['exitcode'] == 1
 
 
@@ -186,6 +205,23 @@ def test_report_crash_and_traceback(tests):
     assert call['traceback'] == traceback
 
 
+def test_report_traceback_styles(make_json):
+    """Handle different traceback styles (`--tb=...`)."""
+    code = '''
+        def test_raise(): assert False
+        def test_raise_nested(): f = lambda: g; f()
+    '''
+    for style in ('long', 'short'):
+        data = make_json(code, ['--json-report', '--tb=%s' % style])
+        for i in (0, 1):
+            assert isinstance(data['tests'][i]['call']['traceback'], list)
+
+    for style in ('native', 'line', 'no'):
+        data = make_json(code, ['--json-report', '--tb=%s' % style])
+        for i in (0, 1):
+            assert 'traceback' not in data['tests'][i]['call']
+
+
 def test_report_item_deselected(make_json):
     data = make_json("""
         import pytest
@@ -240,6 +276,30 @@ def test_report_streams(tests):
     assert 'stderr' not in tests['pass']['call']
 
 
+def test_record_property(make_json, num_processes):
+    data = make_json("""
+        def test_record_property(record_property):
+            record_property('foo', 42)
+            record_property('bar', ['baz', {'x': 'y'}])
+            record_property('foo', 43)
+            record_property(123, 456)
+
+        def test_record_property_empty(record_property):
+            assert True
+
+        def test_record_property_unserializable(record_property):
+            record_property('foo', b'somebytes')
+    """)
+    tests_ = tests_only(data)
+    assert tests_['record_property']['user_properties'] == \
+        [{'foo': 42}, {'bar': ['baz', {'x': 'y'}]}, {'foo': 43}, {'123': 456}]
+    assert 'user_properties' not in tests_['record_property_empty'].keys()
+    # TODO Relay warnings correctly when using xdist
+    if num_processes == 0:
+        assert len(data['warnings']) == 1 and (
+            'not JSON-serializable' in data['warnings'][0]['message'])
+
+
 def test_json_metadata(make_json):
     data = make_json("""
         def test_metadata1(json_metadata):
@@ -358,7 +418,10 @@ def test_warnings(make_json, num_processes):
     assert set(warnings[0]) == {
         'category', 'filename', 'lineno', 'message', 'when'
     }
-    assert warnings[0]['category'] == 'PytestCollectionWarning'
+    assert warnings[0]['category'] in (
+        'PytestCollectionWarning',
+        'PytestWarning'
+    )
     assert warnings[0]['filename'].endswith('.py')
     assert warnings[0]['lineno'] == 1
     assert warnings[0]['when'] == 'collect'
@@ -488,15 +551,15 @@ def test_xdist(make_json, match_reports):
 
 def test_bug_31(make_json):
     data = make_json('''
-        import pytest
+        from flaky import flaky
 
         FLAKY_RUNS = 0
 
-        @pytest.mark.flaky
+        @flaky
         def test_flaky_pass():
             assert 1 + 1 == 2
 
-        @pytest.mark.flaky
+        @flaky
         def test_flaky_fail():
             global FLAKY_RUNS
             FLAKY_RUNS += 1
@@ -529,3 +592,18 @@ def test_bug_41(misc_testdir):
     """#41: Create report file path if it doesn't exist."""
     misc_testdir.runpytest('--json-report', '--json-report-file=x/report.json')
     assert (misc_testdir.tmpdir / 'x/report.json').exists()
+
+
+def test_bug_69(testdir):
+    """#69: Handle deselection of test items that have not been collected."""
+    fn = testdir.makepyfile('''
+        def test_pass():
+            assert True
+        def test_fail():
+            assert False
+    ''').strpath
+    assert testdir.runpytest('--json-report', fn).ret == 1
+    # In this second run, `--last-failed` causes `test_pass` to not be
+    # *collected* but still explicitly *deselected*, so we assert there is no
+    # internal error caused by trying to access the collector obj.
+    assert testdir.runpytest('--json-report', '--last-failed', fn).ret == 1
